/* $Id: log.c,v 1.12 2002/04/29 22:34:41 leemh Exp $
 * -------------------------------------------------------
 * Copyright (c) 2002 Lee Hardy <lee@leeh.co.uk>
 * -------------------------------------------------------
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <string.h>
#include "channels.h"
#include "log.h"
#include "muh.h"
#include "messages.h"

static struct LogHash *log_table[MAX_CHANNELS];
static struct LogHash *global_log_table[MAX_CHANNELS];
int global_logtype;

static void open_logfiles_global();
static char *get_logtimestamp();
static struct LogHash *find_log(char *channel, int logtype);

/* add_log()
 *
 * called from config file parser, creates a logging entry in the hashtable
 */
void add_log(char *log_channel, char *logfile, int logtype)
{
    char *channel;
    int multi = 0;

    /* damn people playing with us! :) */
    if(!log_channel)
	return;

    /* the "global" logentry, mark the logtype and then open logfiles
     * for channels we dont have specific entries for
     */
    if(xstrcmp(log_channel, "*") == 0)
    {
	global_logtype = logtype;
	open_logfiles_global();
	return;
    }

    /* if we're doing multi channels, mark as such so we dont use a 
     * specified logfile, revert to #channel.log
     */
    if((strchr(log_channel, ',')))
	multi = 1;

    for(channel = strtok(log_channel, ","); channel; channel = strtok(NULL, ","))
    {
        struct LogHash *ptr;
        struct LogHash *temp_ptr;
        struct LogHash *last_ptr = NULL;
        unsigned int hashval;

        hashval = hash_channel(channel);

        for(temp_ptr = log_table[hashval]; temp_ptr; temp_ptr = temp_ptr->next)
        {
	    /* adding a logfile for the same channel twice? bzzzzzzt! */
	    if(xstrcasecmp(channel, temp_ptr->channel) == 0)
	        return;

  	    last_ptr = temp_ptr;
        }

        ptr = (struct LogHash *)xcalloc(1, sizeof(struct LogHash));

        ptr->channel = strdup(channel);
        ptr->logtype = logtype;

        /* if were not given a logfile, create one, xmalloc it and as this has
         * its own memory, we can point ptr->filename to it
         */
        if(!logfile || multi)
        {
            logfile = xmalloc(strlen(channel) + 5);
            sprintf(logfile, "%s.log", channel);
   	    ptr->filename = logfile;
        }
        else
            ptr->filename = strdup(logfile);

        if(last_ptr)
            last_ptr->next = ptr;
        else
    	    log_table[hashval] = ptr;

        open_logfile(ptr);
    }
}

/* add_log_global()
 *
 * adds a logentry to the global hashtable
 */
void add_log_global(char *channel)
{
    struct LogHash *ptr;
    struct LogHash *temp_ptr;
    struct LogHash *last_ptr = NULL;
    unsigned int hashval;
    char *logfile;

    hashval = hash_channel(channel);

    for(temp_ptr = global_log_table[hashval]; temp_ptr; temp_ptr = temp_ptr->next)
    {
	/* dont add one twice..! */
	if(xstrcasecmp(channel, temp_ptr->channel) == 0)
	    return;

	last_ptr = temp_ptr;
    }
    
    ptr = (struct LogHash *)xcalloc(1, sizeof(struct LogHash));

    ptr->channel = strdup(channel);
    ptr->logtype = global_logtype;

    logfile = xmalloc(strlen(channel) + 5);
    sprintf(logfile, "%s.log", channel);
    ptr->filename = logfile;

    if(last_ptr)
	last_ptr->next = ptr;
    else
	global_log_table[hashval] = ptr;

    open_logfile(ptr);
}

    
/* rem_log()
 *
 * closes a specific logfile entry
 */
void rem_log(char *channel)
{
    struct LogHash *ptr;
    struct LogHash *next_ptr = NULL;
    unsigned int hashval;

    hashval = hash_channel(channel);

    for(ptr = log_table[hashval]; ptr; ptr = ptr->next)
    {
	if(xstrcasecmp(channel, ptr->channel) == 0)
	{
	    if(ptr->logfile)
	    {
		write_logentry(ptr, LOGM_LOGCLOSE, get_logtimestamp());
	        fflush(ptr->logfile);
		fclose(ptr->logfile);
		ptr->logfile = NULL;

		return;
	    }
        }
    }

    /* search the global log table, if there is one.. */
    if(!global_logtype)
	return;

    for(ptr = global_log_table[hashval]; ptr; ptr = next_ptr)
    {
	next_ptr = ptr->next;

	/* if its a global log entry, remove it from the hashtable
	 * completely, as the only items in there are for active logs
	 */
	if(xstrcasecmp(channel, ptr->channel) == 0)
	{
	    if(ptr->logfile)
	    {
		write_logentry(ptr, LOGM_LOGCLOSE, get_logtimestamp());
		fclose(ptr->logfile);
	    }
	    
	    FREE(ptr->channel);
	    FREE(ptr->logfile);
	    FREE(ptr);
	}
    }
}

/* clear_logs()
 *
 * closes all logfiles and clears hashtables
 */
void clear_logs()
{
    struct LogHash *ptr;
    struct LogHash *next_ptr;
    int i;

    for(i = 0; i < MAX_CHANNELS; i++)
    {
	/* walk the specific entry hashtable */
	for(ptr = log_table[i]; ptr; ptr = next_ptr)
	{
	    next_ptr = ptr->next;

	    if(ptr->logfile)
	    {
	        write_logentry(ptr, LOGM_LOGCLOSE, get_logtimestamp());
	        fclose(ptr->logfile);
	    }

	    xfree(ptr->channel);
	    xfree(ptr->filename);
	    xfree(ptr);
	}

	/* walk the global entry hashtable */
	for(ptr = global_log_table[i]; ptr; ptr = next_ptr)
	{
	    next_ptr = ptr->next;

	    if(ptr->logfile)
	    {
		write_logentry(ptr, LOGM_LOGCLOSE, get_logtimestamp());
		fclose(ptr->logfile);
	    }

	    xfree(ptr->channel);
	    xfree(ptr->filename);
	    xfree(ptr);
	}

	log_table[i] = NULL;
	global_log_table[i] = NULL;
    }
}

/* close_logs()
 *
 * closes all logfiles without clearing hashtable
 */
void close_logs()
{
    struct LogHash *ptr;
    struct LogHash *next_ptr;
    int i;

    for(i = 0; i < MAX_CHANNELS; i++)
    {
        for(ptr = log_table[i]; ptr; ptr = next_ptr)
        {
            next_ptr = ptr->next;

            if(ptr->logfile)
            {
                write_logentry(ptr, LOGM_LOGCLOSE, get_logtimestamp());
                fclose(ptr->logfile);
                ptr->logfile = NULL;
            }
        }

	/* if were closing all logs, this hash needs nuking as it only
	 * contains entries for currently opened logfiles
	 */
	for(ptr = global_log_table[i]; ptr; ptr = next_ptr)
	{
	    next_ptr = ptr->next;

	    if(ptr->logfile)
	    {
		write_logentry(ptr, LOGM_LOGCLOSE, get_logtimestamp());
		fclose(ptr->logfile);
	    }

	    xfree(ptr->channel);
	    xfree(ptr->filename);
	    xfree(ptr);
	}

	global_log_table[i] = NULL;
    }
}

/* open_logfile()
 *
 * opens a logfile if a matching channel is found, based on log hashtable
 */
void open_logfile(struct LogHash *ptr)
{
    if((find_channel(ptr->channel)) >= 0)
    {
	ptr->logfile = fopen(ptr->filename, "a");
	write_logentry(ptr, LOGM_LOGOPEN, get_logtimestamp());
    }
}

/* open_logfiles_global()
 *
 * opens logfiles for all channels we're currently in
 */
void open_logfiles_global()
{
    struct LogHash *ptr;
    int i;

    /* cant open anything if we have no channels */
    if(!channels.amount)
	return;

    /* or if we dont have anything to log.. */
    if(!global_logtype)
	return;

    for(i = 0; i < channels.amount; i++)
    {
        if(!channels.data[i])
	    continue;

	/* a specific logentry, they take precedence */
	if((ptr = find_log(channels.data[i], 0)))
            continue;

	add_log_global(channels.data[i]);
    }
}
	
/* find_log()
 *
 * searches for a log entry in the hashtable
 */
struct LogHash *find_log(char *channel, int logtype)
{
    unsigned int hashval;
    struct LogHash *ptr;

    if(!channel)
        return NULL;

    hashval = hash_channel(channel);

    for(ptr = log_table[hashval]; ptr; ptr = ptr->next)
    {
	if((xstrcasecmp(channel, ptr->channel) == 0) &&
	   ((ptr->logtype & logtype) || !logtype))
	    return ptr;
    }

    return NULL;
}

/* find_log_any()
 *
 * searches for a log entry in the hashtable and global hashtable
 */
struct LogHash *find_log_any(char *channel, int logtype)
{
    unsigned int hashval;
    struct LogHash *ptr;

    if(!channel)
	return NULL;

    /* check the normal hashtable first */
    if((ptr = find_log(channel, logtype)))
	return ptr;

    hashval = hash_channel(channel);

    for(ptr = global_log_table[hashval]; ptr; ptr = ptr->next)
    {
	if((xstrcasecmp(channel, ptr->channel) == 0) &&
	   ((ptr->logtype & logtype) || !logtype))
	    return ptr;
    }

    return NULL;
}

/* write_logentry()
 *
 * writes a logging entry to the logfile
 */
void write_logentry(struct LogHash *ptr, char *format, ...)
{
    char buffer[1024];
    va_list va;

    va_start(va, format);
    vsnprintf(buffer, 1023, format, va);
    va_end(va);

    fprintf(ptr->logfile, "%s", buffer);
    fflush(ptr->logfile);
}

/* write_logentry_all()
 *
 * writes a logging entry to all matching logfiles
 */
void write_logentry_all(int type, char *format, ...)
{
    struct LogHash *ptr;
    char buffer[1024];
    va_list va;
    int i;

    va_start(va, format);
    vsnprintf(buffer, 1023, format, va);
    va_end(va);

    for(i = 0; i < channels.amount; i++)
    {
	if(channels.data[i] && (ptr = find_log_any(channels.data[i], type)))
	    write_logentry(ptr, "%s", buffer);
	
    }
}

/* get_logtimestamp()
 *
 * creates a timestamp in the form:
 *    Sun Jan 02 11:53:33 2002
 */
char *get_logtimestamp()
{
    time_t now;
    struct tm *form;
    static char stamp[100];

    time(&now);
    form = localtime(&now);
    strftime(stamp, 99, "%a %b %d %H:%M:%S %Y", form);
    return stamp;
}

